/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
/*******************************************************/
/* file: ports.c                                       */
/* abstract:  This file contains the routines to       */
/*            output values on the JTAG ports, to read */
/*            the TDO bit, and to read a byte of data  */
/*            from the prom                            */
/*                                                     */
/* Notes: See the notes for micro.c for explanation of */
/*        the compiler switch  "DEBUG_MODE".           */
/*******************************************************/
#if defined( DEBUG_MODE ) || defined( WIN95PP ) || defined( WINDRVR )
#include "stdio.h"
#include "xcommon.h"
#include "x_bsdl.h"
#endif

#ifdef DEBUG_MODE
/* if in debugging mode, use variables instead of setting the ports */
short pTCK,pTMS,pTDI;
#endif

#ifdef WINDRVR
#include "bjdrvif.h"
jdriver *JD = NULL;
static int port = 0;
#endif

#if defined( WIN95PP ) || defined( WINDRVR )
#define DELAY_SAMPLES 10
#define DELAY_CHECK_LOOPS 10000
#include <conio.h>
#include <windows.h>
/* delay count for one millisecond delay */
static int one_ms_delay = 0;
static inPortType in_word;
static outPortType out_word;
static void delay_loop(int count);
static void calibrate_delay( void );
static long get_tick_count( void );
#endif

#ifdef WIN95PP
static unsigned short base_port = 0x378;
#endif

extern int iGlobParportPin17Setting,iDbg_PrintBsdl;
//==!========================================================================
//==! 2002/03/25: Randal's parallel cable 3 burst mode code.
//==!========================================================================
//==! 2002/04/08: Added dwCtrlPort and associated code to set bit 3 of the
//==! parallel port control register. This was necessary to make the program
//==! work with the PC-4 cable which uses this bit to enable or disable the
//==! CoolRunner device in the cable. Not setting bit 3 caused the PC-4 cable
//==! to read back incorrect data. This seems to be necessary when the code
//==! only uses the burst mode function exclusively. Randal's code, which
//==! mixes calls to setPort() and WD_MultiTransfer(), does not require
//==! explicitly setting control reg bit 3. This code works with both the
//==! PC-3 and PC-4 cables.
//==!
//==! 2002/03/28: MAX_BIT_BURST_SIZE (max bit burst shift size) used by
//==! burst_pTransfer data structure (64K was arbitrarily selected).
//==! burst_pTransfer[] will hold data bit and command to be used by the port
//==! driver for that bit.
//==!
//==! To shift out MAX_BYTE_BURST_SIZE of 64K bytes (64*1024)*(8 bits/byte)
//==! requires at most 4 entries:
//==! - 1 to write the TDI and TMS data
//==! - 1 to read the TDO data from the port
//==! - 1 to write/drive TCK high
//==! - 1 to write/drive TCK back low
//==!========================================================================
DWORD       burst_dwDataPort     = 0;
DWORD       burst_dwStatusPort   = 0;
DWORD       dwCtrlPort           = 0;
// BOOL        burst_bSimulate      = 0;
BOOL        burst_bInit          = 0;
WD_TRANSFER* burst_pTransfer     = 0;

unsigned char   ucCtrlByte       = 0;

void xportsInitialize()
{
  unsigned char   ucPortStatusOn  = 0;
  unsigned char   ucPortStatusOff = 0;
  int             port            = 0;

  unsigned char   ucPortState     = 0;  //==! 2003/01/17 CSD-RAD
  unsigned char   ucDataVal       = 0;
  int             iXportVal;

    /* Initialize ports and calibrate timers (if necessary) */
  calibrate_delay();
  out_word.bits.one = 1;
  out_word.bits.zero = 0;

  JD = (jdriver *) new jdriver;
  iXportVal = JD->GetParallelPorts( port );
  if (iXportVal == -1)
  { delete JD;
    JD = NULL;
    dispInfo(0,
"Error: Invalid parallel port address detected. Install or start Xilinx\n");
    dispInfo(0,
"       Windriver device driver. To install driver, download and run\n");
    dispInfo(0,
"       ftp.xilinx.com/pub/utilities/M1_pc/pc3driver.exe, or search for\n");
    dispInfo(1,
"       \"Answer record 13360\" in AnswerDatabase in support.xilinx.com.\n");
  } // IF

  if (iOpt_ProgDebug == 1)
  { sprintf(sMsgBuf00,"Using parallel port addr [0x%03x]\n",iXportVal);
    dispInfo(0,sMsgBuf00);
  } // IF

  burst_dwDataPort     = JD->GetDataPort();
  burst_dwStatusPort   = JD->GetStatusPort();
  dwCtrlPort           = JD->GetControlPort();

//==! 2003/01/28: Put part into PS2 mode
  JD->WriteByte(0x77A,0x34);

//==! 2003/01/17: Display parallel port status for debugging
  ucPortState  = (unsigned char) JD->ReadByte( burst_dwStatusPort );

  JD->WriteByte( burst_dwDataPort, 0x40 );
  ucPortStatusOn  = (unsigned char) JD->ReadByte( burst_dwStatusPort );

  JD->WriteByte( burst_dwDataPort, 0x00 );
  ucPortStatusOff = (unsigned char) JD->ReadByte( burst_dwStatusPort );

  ucCtrlByte = JD->ReadByte(dwCtrlPort);
  ucCtrlByte &= ~(0x20);  //==! Clear bit 1
  ucCtrlByte &= ~(0x08);  //==! Clear bit 3 to drive parport pin 17 high
  JD->WriteByte(dwCtrlPort,ucCtrlByte);
  ucCtrlByte |= 0x08;     //==! Set bit 3 to drive parport pin 17 low
  JD->WriteByte(dwCtrlPort, ucCtrlByte);

#if 0
  if ( ( ( ucPortStatusOn & 0x20 ) != 0x20 ) || ( ucPortStatusOff & 0x20 ) )
  { printf( "ERROR:  Xilinx Parallel Cable III is not connected to parallel port.\n" );
    burst_bSimulate  = TRUE;
  }
#endif

  ucPortStatusOff = (unsigned char) JD->ReadByte( burst_dwStatusPort );
//==!========================================================================
//==! 2003/01/28: From detectCable() code by Jun Flores (NOTE_0002)
//==!========================================================================
  ucPortStatusOff |= BUSY_PIN_MASK;
  ucPortStatusOff &= ~PE_PIN_MASK;

  if ((ucPortStatusOff & BUSY_PIN_MASK) &&
      ((ucPortStatusOff & PE_PIN_MASK) == 0))
  { //==! Set the following bits (counting from 0):
    //==! d3-set to zero to enable cable output drivers
    //==! d4-set to one to be able to read tdo,d/p
    //==! d6-set to one to test pe and busy lines
    ucDataVal = 0x50;
    JD->WriteByte(burst_dwDataPort,ucDataVal);
    ucPortState = JD->ReadByte (burst_dwStatusPort);
    ucPortState |= ACK_PIN_MASK; // force D6 to 1
    ucPortState |= PE_PIN_MASK;
    ucPortState &= ~BUSY_PIN_MASK;
    if (((ucPortState & BUSY_PIN_MASK) == 0) && (ucPortState & PE_PIN_MASK))
    { // check for PC4/HW-140, ACK should follow d6
      if (ucPortState & ACK_PIN_MASK)
      { ucDataVal &= ~(0x48); // reset bit 3 and 6 (bit 3 = 0 will enable
                              // cable output drivers)
        JD->WriteByte(burst_dwDataPort,ucDataVal);
      } // IF
    } // IF
    else
    { printf("Err: detectCable Error2: BUSY pin NOT 0 or PE pin NOT 1\n");
    } // ELSE
  } // IF
//==!========================================================================

  burst_pTransfer  = new WD_TRANSFER[ MAX_BIT_BURST_SIZE ];
  ZeroMemory(burst_pTransfer,( MAX_BIT_BURST_SIZE * sizeof( WD_TRANSFER ) ));

  if (iGlobParportPin17Setting == 1)
  { ucCtrlByte &= ~(0x08);  //==! Clear bit 3 to drive parport pin 17 high
    JD->WriteByte(dwCtrlPort,ucCtrlByte);
  } // IF

}
//==!========================================================================

void closePort()
{
#ifdef WINDRVR
//==! Set bit 3 to 1 to set parallel port pin 17 to 0
  if (iGlobParportPin17Setting == 1 && JD != NULL)
  { ucCtrlByte |= 0x08;
    JD->WriteByte(dwCtrlPort,ucCtrlByte);
  } // IF

  if (JD != (jdriver *) NULL)
  { JD->close();
    delete JD;
    JD = NULL;
  }

  if (burst_pTransfer != NULL)
  { delete [] burst_pTransfer;
    burst_pTransfer = NULL;
  } // IF

#endif
} /* closePort() */

//==!========================================================================
int setPort(short p,short val)
{
#ifdef DEBUG_MODE
  if (p==TCK)
    pTCK=val;
  if (p==TMS)
    pTMS=val;
  if (p==TDI)
    pTDI=val;
#endif

#if defined(WIN95PP) || defined(WINDRVR)
  if (p==TMS)
    out_word.bits.tms = (unsigned char) val;
  if (p==TDI)
    out_word.bits.tdi = (unsigned char) val;
  if (p==TCK)
  { out_word.bits.tck = (unsigned char) val;
#ifdef WINDRVR
    JD->WriteByte( JD->GetDataPort(), out_word.value );
#else
    (void) _outp( (unsigned short) (base_port + 0), out_word.value );
#endif
  }
#endif

#if 1
  return 0;
#endif

}


//==!========================================================================
#ifdef DEBUG_MODE
void printPorts()
{
	printf("M=%d  I=%d\n",pTMS,pTDI);
}
#endif


//==!========================================================================
/* toggle tck LHL */
unsigned char pulseClock( int getTdo )
{
    unsigned char   tdoVal  = 0;
	setPort(TCK,0);  /* set the TCK port to low  */
    if ( getTdo )
    {
        tdoVal  = readTDOBit();
    }
	setPort(TCK,1);  /* set the TCK port to high */
#ifdef DEBUG_MODE
	/* if in debugging mode, print the ports on the rising clock edge */
	printPorts();
#endif
	/*setPort(TCK,0);  /* set the TCK port to low  */
    return( tdoVal );
}


//==!========================================================================
/* read the TDO bit from port */
unsigned char readTDOBit()
{
#ifdef DEBUG_MODE
	static int i = 0;
	if (i++ < 6)
		return 0;
	else {
		i = 0;
		return 1;  /* garbage value for now; replace with real port read. */
	}
#endif

#if defined( WIN95PP ) || defined( WINDRVR )
#ifdef WINDRVR
	in_word.value = (unsigned char) JD->ReadByte( JD->GetStatusPort() );
#else
	in_word.value = (unsigned char) _inp( (unsigned short) (base_port + STATUS_OFFSET) );
#endif
	if (in_word.bits.tdo == 0x1) {
		return( (unsigned char) 1 );
	}
	return( (unsigned char) 0 );
#endif
}

//==!========================================================================
/* Wait at least the specified number of microsec.                           */
/* Use a timer if possible; otherwise estimate the number of instructions    */
/* necessary to be run based on the microcontroller speed.  For this example */
/* we pulse the TCK port a number of times based on the processor speed.     */
//==!========================================================================
//==! 2003/01/20: Routine calls waitUsec() for microsecond resolution timing
//==! and ::Sleep() for millisecond resolution timing. This routine should
//==! be customized for other embedded systems.
//==!========================================================================
void waitTime(long microsec)
{
#if defined( DEBUG_MODE ) || defined( WIN95PP ) || defined( WINDRVR )

  int iMillisecDelay;

  if (iOpt_MinOneClk == 1)
  { pulseClock(0);
    setPort(TCK,0);  //==! set the TCK port to low
  } // IF

  if (microsec < 1000)
  { waitUsec(microsec);
  } // IF
  else
  {
// Calculate millisecond wait time.  Add 999 to do ceil on the int divide.
    iMillisecDelay  = ( ( microsec + 999 ) / 1000);

    if ( iMillisecDelay )
    { if ( g_verbose > 1 )
      { printf( "Wait in Run-Test/Idle:  %d msec.\n", iMillisecDelay );
      }
      ::Sleep( iMillisecDelay );
    } // IF
  } // ELSE
#else
  long repeat;

#define CLOCK_RATE 150 /* set to be the clock rate of the system in kHz */
  long clockRunTests=microsec*CLOCK_RATE/1000;

  for (repeat=0;repeat<clockRunTests;repeat++)
  { pulseClock( 0 );
  } // FOR
#endif
} // waitTime()

/*==!===================================================================!==*/
/*==! Prepare data for burst mode output.                                  */
/*==!                                                                      */
/*==! Function Parameters : iMode = if 0, then add TMS and TDI data to the */
/*==!                           burst data structure; if 1, then do a data */
/*==!                           burst if there is data in the structure    */
/*==!                       iByteBoundary = if 1, then function should     */
/*==!                           read back a full set of bits for each byte */
/*==!                           (i.e., should read back 16 bits or 2 bytes */
/*==!                           and not 15 bits)                           */
/*==!                       iReadTDO = if 1 then read TDO from port, else  */
/*==!                           ignore port status/read bit                */
/*==!                       plBurstDataIndex = pointer to variable that    */
/*==!                           keeps track of index into burst data       */
/*==!                           structure                                  */
/*==!                       iTms = TMS data                                */
/*==!                       uTdi = TDI data                                */
/*==!                       plByteCount = pointer to byte count variable   */
/*==!                       plReadCnt = pointer to long var which records  */
/*==!                           number of bits read back                   */
/*==!                       pcArrayRead = pointer to array of chars to     */
/*==!                           hold burst read data bytes                 */
/*==! Return Value        : 1 if data was sent out in a burst, else 0 to   */
/*==!                       indicate that data is being collected for      */
/*==!                       a later data burst.                            */
/*==!===================================================================!==*/
int burstData(int iMode,int iByteBoundary,int iReadTDO,
      long *plBurstDataIndex,
      int iTms,int iTdi,long *plReadCnt,char *pcArrayRead)
{ long lVar0;
  int iBitCnt;
  char cReadByte;

  *plReadCnt = 0;

  if (iMode == 1)
  { if (*plBurstDataIndex > 0)
    { WD_MultiTransfer(JD->GetHWD(),burst_pTransfer,*plBurstDataIndex);
      if (iReadTDO == 1)
      { if (*plBurstDataIndex%8 != 0 && iByteBoundary == 1)
        { dispInfo(1,"DataBurst != multiple of 8\n");
        } // IF
        *plReadCnt = 0;
        for (lVar0 = 1,cReadByte = 0,iBitCnt = 0;
             lVar0 < *plBurstDataIndex; lVar0 += 4)
        { in_word.value = burst_pTransfer[lVar0].Data.Byte;
          cReadByte = cReadByte | (in_word.bits.tdo << iBitCnt);
          ++iBitCnt;
          if (iBitCnt >= 8)
          { pcArrayRead[*plReadCnt] = cReadByte;
            ++(*plReadCnt);  //==! Increment number of bytes read
            iBitCnt = 0;
            cReadByte = 0;
          } // IF iBitCnt >= 8
        } // FOR lVar0 < *plBurstDataIndex
        if (iBitCnt > 0)
        { pcArrayRead[*plReadCnt] = cReadByte;
          ++(*plReadCnt);  //==! Increment number of bytes read
        } // IF
      } // IF iReadTDO == 1
    } // IF *plBurstDataIndex > 0
    *plBurstDataIndex = 0;
    return 1;
  } // IF iMode == 1

  out_word.bits.tdi  = iTdi;
  out_word.bits.tms  = iTms;
  out_word.bits.tck  = 0;
  burst_pTransfer[*plBurstDataIndex].cmdTrans   = WP_BYTE;
  burst_pTransfer[*plBurstDataIndex].dwPort     = burst_dwDataPort;
  burst_pTransfer[*plBurstDataIndex].Data.Byte  = out_word.value;
  ++(*plBurstDataIndex);

  if (iReadTDO == 1)
  { burst_pTransfer[*plBurstDataIndex].cmdTrans = RP_BYTE;
    burst_pTransfer[*plBurstDataIndex].dwPort = burst_dwStatusPort;
    burst_pTransfer[*plBurstDataIndex].Data.Byte = 0x0;
    ++(*plBurstDataIndex);
  }

  out_word.bits.tck                      = 1;
  burst_pTransfer[*plBurstDataIndex].cmdTrans   = WP_BYTE;
  burst_pTransfer[*plBurstDataIndex].dwPort     = burst_dwDataPort;
  burst_pTransfer[*plBurstDataIndex].Data.Byte  = out_word.value;
  ++(*plBurstDataIndex);
  out_word.bits.tck                      = 0;
  burst_pTransfer[*plBurstDataIndex].cmdTrans   = WP_BYTE;
  burst_pTransfer[*plBurstDataIndex].dwPort     = burst_dwDataPort;
  burst_pTransfer[*plBurstDataIndex].Data.Byte  = out_word.value;
  ++(*plBurstDataIndex);

//==! Burst the data/command information when this limit is reached to avoid
//==! overrunning the array. Note that this code section should never be
//==! reached when using this function to read back data. The calling
//==! function should explicitly burst the data (by setting iMode = 1) and
//==! then reading the data back from appropriate array locations in
//==! burst_pTransfer[]. This should only be executed when bursting data out
//==! without expecting any data to be read back.
  if (*plBurstDataIndex > (MAX_BIT_BURST_SIZE - 6))
  { WD_MultiTransfer(JD->GetHWD(),burst_pTransfer,*plBurstDataIndex);
    if (iReadTDO == 1)
    { dispInfo(1,"iReadTdo = 1");
    } // IF
    *plBurstDataIndex = 0;
    return 1;
  } // IF

  return 0;

} // burstData()
/*==!======================================================================*/


#if defined(WIN95PP) || defined(WINDRVR)
/************************************************************************
*
*	get_tick_count() -- Get system tick count in milliseconds
*
*/
static long get_tick_count(void)
{
    long tick_count = 0L;
	tick_count = GetTickCount();
    return (tick_count);
}

static void calibrate_delay(void)
{
	int sample = 0;
	int count = 0;
	long tick_count1 = 0L;
	long tick_count2 = 0L;

	one_ms_delay = 0;

	for (sample = 0; sample < DELAY_SAMPLES; ++sample)
	{
		count = 0;
		tick_count1 = get_tick_count();
		while ((tick_count2 = get_tick_count()) == tick_count1) {};
		do { delay_loop(DELAY_CHECK_LOOPS); count++; } while
			((tick_count1 = get_tick_count()) == tick_count2);
		one_ms_delay += (int)((DELAY_CHECK_LOOPS * (long)count) /
			(tick_count1 - tick_count2));
	}

	one_ms_delay /= DELAY_SAMPLES;
}

static void delay_loop(int count)
{
	while (count != 0) count--;
}
#endif
//==!========================================================================
//==! End of PORTS.CPP
//==!========================================================================
