/***************************************************************************/
/* J Drive (tm) Engine                                                     */
/* Copyright 2003-2005, Xilinx, Inc.                                       */
/* The use, reproduction, and distribution of this code is subject         */
/* to the J Drive license requirements and restrictions.                   */
/***************************************************************************/
//==!=====================================================================!==
//==! See XPROJECT.H for project file list and project version history.
//==!=====================================================================!==
//==! Revision history: XPARSER.CPP
//==! 2003/03/30: Data parser routines
//==!=====================================================================!==
#include "xcommon.h"
#include "xproject.h"

//==! User should supply the following functions to parse command line for
//==! program specific options.
extern int processProgOptions(int *argc,char ***argv);
extern  void displayProgOptions();
extern int checkOptCombo();
extern int iOpt_SkipIdCheck,iOpt_PrintBsdl,iOpt_ProgDebug,iOpt_FlowLoopDbg,
   iOpt_RptCntDispOff,iGlobOptCheckBsdl;
//==!=====================================================================!==
//==! Reset/initialize global program variables that are used when the
//==! program starts, and that are also used when the program is reinvoked as
//==! part of a loop (i.e., when processing another input file).
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void resetGlobalVars()
{
  iOptOverWritePrompt = 1;
  iOptShowStatus = 1;
  iOpt_FlowLoopDbg = 0;
  lSpecOptions = 0;

} // resetGlobalVars()
//==!=====================================================================!==
//==! Initialize global program variables. This routine should only be called
//==! once but resetGlobalVars() can be called multiple times to reprocess
//==! the same command line options without restarting the program.
//==!
//==! Function Parameters: none
//==! Return Values      : none
//==!=====================================================================!==
void initGlobalVars()
{

  t_ProgStart = 0;
  t_ProgFinish = 0;

  pDP_Input = NULL;
  pDP_Output = NULL;
  pDP_Log = NULL;
  pDP_List = NULL;

  iGlobOptBufferedReadOp = 1;
  lGlobOptBufferedReadSize = dREADBUFMAX;
  iGlobOptBufferedWriteOp = 1;
  lGlobOptBufferedWriteSize = dWRITEBUFMAX;

  resetGlobalVars();

} // initGlobalVars()
//==!=====================================================================!==
//==! Check if pointer is NULL.
//==!
//==! Function Parameters: pvPtr = void pointer
//==!                      sMsg = error message (can be NULL)
//==! Return Value       : none
//==!=====================================================================!==
void isPtrNull(void *pvPtr,char *sMsg)
{

  if (pvPtr == NULL)
  { sprintf(sMsgOnly,"Null [%s]",sMsg != NULL ? sMsg : "*NoMsg*");
    jMessage(DISP_AND_EXIT,sMsgOnly,0,NULL);
  } // IF pvPtr == NULL

} // isPtrNull()
//==!=====================================================================!==
//==! Check if lVal is in range [lMinV .. lMaxV] inclusive. Return 1 if in
//==! range. If out of range and:
//==!
//==! - iMode = 0 then return 0
//==!         = 1 then call jMessage()
//==!
//==! Function Parameters: iMode = see function description
//==!                      lVal  = val to test
//==!                      lMinV = min of range
//==!                      lMaxV = max of range
//==!                      sMsg  = for jMessage()
//==!                      iStat = for jMessage()
//==! Return Value       : 1 if in range, 0 otherwise
//==!=====================================================================!==
int isInRange(int iMode,long lVal,long lMinV,long lMaxV,char *sMsg,int iStat)
{

  if (lVal < lMinV || lVal > lMaxV)
  { if (iMode != 0)
    { sprintf(sMsgOnly,"[%s] range[%ld <= %ld <= %ld]",
              sMsg,lMinV,lVal,lMaxV);
      jMessage(iStat,sMsgOnly,0,NULL);
    } // IF
    return 0;
  } // IF
  return 1;

} // isInRange()
//==!=====================================================================!==
//==! Free memory used by global data ports and pointers. Other program or
//==! application-specific variables that allocated memory are deleted by
//==! freeApplicationMem().
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void freeGlobalMem()
{ cDataPort *pDP;

//==! Free memory allocated for ports in the linked list.
  while (pDP_List != NULL)
  { pDP = pDP_List->pNextDataPort;
    delete pDP_List;
    pDP_List = pDP;
  } // WHILE

  while (pDP_Input != NULL)
  { pDP = pDP_Input->pNextDataPort;
    delete pDP_Input;
    pDP_Input = pDP;
  } // WHILE

} // freeGlobalMem()
//==!=====================================================================!==
//==! Send message to stderr according to uiDispInstr.
//==!
//==! Function Parameters: sMsg = message
//==!                      uiDispInstr = display instructions
//==!                      iMode: if 1 then display sMsg
//==!                      iLogMode: if 1 then send sMsg to pDP_Log
//==! Return Value       : none
//==!=====================================================================!==
void dispMsg(char *sMsg,unsigned int uiDispInstr,int iMode,int iLogMode)
{ int iVar,iTermPos,iCnt;
  FILE *pF_MsgDest;

//==! Do message display routine twice: first iteration will display to
//==! stderr, 2nd iteration will write to global log file (if open).
  for (iCnt = 0; iCnt < 2; ++iCnt)
  { pF_MsgDest = NULL;
    switch (iCnt)
    { case 0:
      { if (iMode == 1 && iOptShowStatus == 1)
        { pF_MsgDest = stderr;
        } // IF
        break;
      } // CASE 0
      case 1:
      { if (pDP_Log != NULL && iLogMode == 1)
        { if (pDP_Log->pFILE == NULL)
          { if (pDP_Log->psPortName == NULL)
            { fprintf(stderr,"\ndispMsg: Log port NULL");
            } // IF
            else
            { pDP_Log->pFILE = xFileOpen(pDP_Log->psPortName,"w");
            } // ELSE
          } // IF
          pF_MsgDest = pDP_Log->pFILE;
        } // IF pDP_Log != NULL
        break;
      } // CASE 1
      default:
      { break;
      } // DEFAULT
    } // SWITCH (iCnt)

    if (pF_MsgDest == NULL)
    { continue;
    } // IF pF_MsgDest == NULL

    iVar = 0;
    iTermPos = 0;

    if ((uiDispInstr & DISP_CLEAR   ) != 0 ||
        (uiDispInstr & DISP_PRE_CR  ) != 0)
    { if (iCnt == 0)
      { fprintf(pF_MsgDest,"\r");
        iVar = 1;
      } // IF iCnt == 0
    } // IF
    if ((uiDispInstr & DISP_PRE_LF  ) != 0)
    { fprintf(pF_MsgDest,"\n");
    } // IF

    if ((uiDispInstr & DISP_CLEAR   ) == 0)
    { iTermPos = fprintf(pF_MsgDest,sMsg);
    } // IF iDispInstr != DISP_CLEAR

    if (iCnt == 0 && iVar == 1)
    { for (iVar = iTermPos; iVar < 79; ++iVar)
      { fprintf(pF_MsgDest," ");
      } // FOR iVar < 79
    } // IF iCnt == 0 && iVar == 1

    if ((uiDispInstr & DISP_CLEAR   ) != 0 ||
        (uiDispInstr & DISP_POST_CR ) != 0)
    { if (iCnt == 0)
      { fprintf(pF_MsgDest,"\r");
      } // IF iCnt == 0
    } // IF
    if ((uiDispInstr & DISP_POST_LF ) != 0)
    { fprintf(pF_MsgDest,"\n");
    } // IF

  } // FOR iCnt < 2

} // dispMsg()
//==!=====================================================================!==
//==! readCmdLineData() returns a pointer to data from the command line given
//==! the current argc and argv values. If iMode = 1, then decrement argc and
//==! advance argv pointer to the next command line data.
//==!
//==! Function Parameters: argc = argument count
//==!                      argv = arguments
//==!                      iMode = see above description
//==! Return Value       : none
//==!=====================================================================!==
char *readCmdLineData(int *argc,char ***argv,int iMode)
{
  if (iMode == 1)
  { --(*argc);
    if (*argc <= 0)
    { return NULL;
    } // IF
    ++(*argv);
  } // IF
  return **argv;

} // readCmdLineData()
//==!=====================================================================!==
//==! Return next character from data port through the cDataPort's
//==! iCurChar variable. If iMode = 0, then return the first char read, else
//==! return the first non-whitespace character.
//==!
//==! Function Parameters: pDP = pointer to cDataPort
//==!                      iMode = see function description
//==! Return Value       : none
//==!=====================================================================!==
void getDpChar(cDataPort *pDP,int iMode)
{

  isPtrNull(pDP->pFILE,"getDPChar");
  if (pDP->iParserGetChar == 1 || iMode == 1)
  { if (iMode == 1 && pDP->iParserGetChar == 0)
    { pDP->iParserGetChar = 1;
      if (!isspace(pDP->iCurChar))
      { return;
      } // IF
    } // IF
    do
    {
//==! 2003/12/31: iGlobOptBufferedReadOp = 1 specifies use of fread() to
//==! buffer read operations from the data port or file. If
//==! iOptBufferedOp = 0, then read data using fgetc().
      if (iGlobOptBufferedReadOp == 0)
      { if (feof(pDP->pFILE))
        { jMessage(DISP_AND_EXIT,"EOF_dp",0,NULL);
        } // IF
        pDP->iCurChar = fgetc(pDP->pFILE);
        if (feof(pDP->pFILE))
        { pDP->iCurChar = ' ';
          pDP->m_iEndOfDataStream = 1;
          return;
        } // IF
      } // IF iGlobOptBufferedReadOp == 0
      else
      { if (pDP->m_lReadBufIndex >= pDP->m_lReadBufMax)
        { if (pDP->m_iBufDataReadEnd == 1)
          { if (pDP->m_iEndOfDataStream == 0)
            { pDP->m_iEndOfDataStream = 1;
              pDP->iCurChar = ' ';
              return;
            } // IF
            jMessage(DISP_AND_EXIT,"EOF_dp buf",0,NULL);
          } // IF
          else
          { pDP->m_lReadBufIndex = 0;
            pDP->m_lReadBufMax = fread(pDP->m_csReadBuf,
                       sizeof(char),lGlobOptBufferedReadSize,pDP->pFILE);
//==! Note that if the data port is a FILE, then end of data condition can be
//==! checked using feof(). If data port is some other source, then use the
//==! appropriate test to set the m_iBufDataReadEnd member variable.
            if (feof(pDP->pFILE))
            { pDP->m_iBufDataReadEnd = 1;
            } // IF
          } // ELSE
        } // IF
        pDP->iCurChar = pDP->m_csReadBuf[pDP->m_lReadBufIndex];
        ++pDP->m_lReadBufIndex;
      } // ELSE

      if (pDP->iCurChar != '\r')
      { ++pDP->lCharCount;
      } // IF

      if (pDP->iCurChar == '\n')
      { ++(pDP->lLineNum);
      } // IF
      if (iMode == 0)
      { return;
      } // IF
    } while (isspace(pDP->iCurChar));
  } // IF iParserGetChar == 1
  pDP->iParserGetChar = 1;

} // getDpChar()
//==!=====================================================================!==
//==! Skip the rest of the cDataPort line by reading and discarding all
//==! characters from the current cDataPort FILE position to the end of
//==! line.
//==!
//==! Function Parameters: pDP = data port pointer
//==! Return Value       : none
//==!=====================================================================!==
void skipDpLine(cDataPort *pDP)
{ 

  do
  { if (pDP->m_iEndOfDataStream == 1)
    { return;
    } // IF
    getDpChar(pDP,0);
  } while (pDP->iCurChar != '\n');

} // skipDpLine()
//==!=====================================================================!==
//==! Create a copy of pSrcStr: allocate memory to pStr and copy pSrcStr.
//==!
//==! Function Parameters: pStr = pointer to char pointer
//==!                      pSrcStr = string to copy
//==! Return Value       : none
//==!=====================================================================!==
void copyStr(char **pStr,char *pSrcStr)
{

  if (pSrcStr != NULL)
  { *pStr = new char [strlen(pSrcStr)+1];
    isPtrNull((void *)(*pStr),"copyStr");
    strcpy(*pStr,pSrcStr);
  } // IF

} // copyStr()
//==!=====================================================================!==
//==! cCharStr constructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
cCharStr::cCharStr()
{
  m_lSize = 0;
  m_csStr = new char [MAX_XDATA_SIZE+1];
  m_csStr[0] = '\0';
  m_pNextcCharStr = NULL;

} // cCharStr::cCharStr()
//==!=====================================================================!==
//==! cCharStr destructor
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
cCharStr::~cCharStr()
{
  m_lSize = 0;
  freeArray(m_csStr);

} // cCharStr::~cCharStr()
//==!=====================================================================!==
//==! Reset port variables; usually done to prepare reading from the
//==! beginning of the data file.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void cDataPort::resetDataPort()
{

  sOrigData[0] = '\0';
  sData[0]     = '\0';
  if (m_pcCharStr != NULL)
  { m_pcCharStr->m_csStr[0] = '\0';
  } // IF

  lLineNum  = 1;
  lCharCount = 0;
  iCurChar = 32;  // Initialize to white-space (ASCII space character)
  iBracketLevel = 0;
  iBsdlEntityEnd = 0;

  iParserWarnings = 0;
  iParserGetData = 1;
  iParserGetChar = 1;

  m_lReadBufIndex = 0;
  m_lReadBufMax = 0;
  m_iBufDataReadEnd = 0;
  m_iEndOfDataStream = 0;

  m_iDataFormat = 0;
  m_iFirstData = 1;

  if (iGlobOptBufferedWriteOp == 1)
  { m_csWriteBuf[0] = '\0';
    m_lWriteBufCnt = 0;
  } // IF

} // cDataPort::resetDataPort()
//==!=====================================================================!==
//==! Reopen the data port.
//==!
//==! Function Parameters: sMode = port open mode
//==! Return Value       : none
//==!=====================================================================!==
void cDataPort::xPortReopen(char *sMode)
{
  xFclose(&pFILE);
  resetDataPort();
  pFILE = xFileOpen(psPortName,sMode);

} // cDataPort::xPortReopen()
//==!=====================================================================!==
//==! Close the data port.
//==!
//==! Function Parameters: none
//==! Return Value       : none
//==!=====================================================================!==
void cDataPort::xCloseDataPort()
{
  xFclose(&pFILE);

} // cDataPort::xCloseDataPort()
//==!=====================================================================!==
//==! Data port constructor: create data port with name psStr.
//==!
//==! Function Parameters: psStr
//==! Return Value       : none
//==!=====================================================================!==
cDataPort::cDataPort(char *psStr)
{

  if (psStr == NULL)
  { psPortName = NULL;
  } // IF
  else
  { copyStr(&psPortName,psStr);
  } // ELSE
  pFILE = NULL;
  iCommentStyle = XDATA_COMMENT_STYLE_BSDL;
  pNextDataPort = NULL;

  m_csReadBuf = NULL;
  if (iGlobOptBufferedReadOp == 1)
  { m_csReadBuf = new char [lGlobOptBufferedReadSize];
  } // IF

  m_csWriteBuf = NULL;
  if (iGlobOptBufferedWriteOp == 1)
  { m_csWriteBuf = new char [lGlobOptBufferedWriteSize];
  } // IF

  sData = new char [MAX_XDATA_SIZE+1];
  sOrigData = new char [MAX_XDATA_SIZE+1];
  m_lDpDataSize = MAX_XDATA_SIZE;

  m_pcCharStr = new cCharStr;

  resetDataPort();

} // cDataPort::cDataPort()
//==!=====================================================================!==
//==! Free memory used by the data port. Set port pointer to NULL after
//==! the delete operation.
//==!
//==! Function Parameters: pDP = pointer to data port pointer
//==! Return Value       : none
//==!=====================================================================!==
cDataPort::~cDataPort()
{

  freeArray(psPortName);
  freeArray(m_csReadBuf);
  freeArray(m_csWriteBuf);
  freeArray(sData);
  freeArray(sOrigData);
  freeData(m_pcCharStr);
  xFclose(&pFILE);

} // cDataPort::~cDataPort()
//==!=====================================================================!==
//==! Write csData to data port. This routine manages the write buffer and
//==! calls fwrite() when it is full or near full.
//==!
//==! Function Parameters: iMode = 0: use csData
//==!                              1: fflush() buffer
//==!                      csData = char string to write
//==! Return Value       : 1 if successful, 0 otherwise
//==!=====================================================================!==
int cDataPort::writeToCDP(int iMode,char *csData)
{

  if (pFILE == NULL)
  { return 0;
  } // IF

  if (iGlobOptBufferedWriteOp == 0)
  { if (iMode == 0)
    { fprintf(pFILE,"%s",csData);
    } // IF
  } // IF iGlobOptBufferedWriteOp == 0
  else
  { switch(iMode)
    { case 0:
      { m_lWriteTmpBufCnt = strlen(csData);
        if ((m_lWriteBufCnt + m_lWriteTmpBufCnt) >
                 lGlobOptBufferedWriteSize-1)
        { fwrite(m_csWriteBuf,sizeof(char),m_lWriteBufCnt,pFILE);
          m_lWriteBufCnt = m_lWriteTmpBufCnt;
          strcpy(m_csWriteBuf,csData);
        } // IF
        else
        { strcat(m_csWriteBuf,csData);
          m_lWriteBufCnt += m_lWriteTmpBufCnt;
        } // ELSE
        break;
      } // CASE 0
      case 1:
      { if (m_lWriteBufCnt > 0)
        { fwrite(m_csWriteBuf,sizeof(char),
                 m_lWriteBufCnt,pFILE);
          if (fflush(pFILE) != 0)
          { jMessage(DISP_SHOW_MSG,"fwrite() + fflush()",0,NULL);
          } // IF
          m_lWriteBufCnt = 0;
        } // IF
        m_csWriteBuf[0] = '\0';
        break;
      } // CASE 1:
      default:  // should never reach this code
      { jMessage(DISP_AND_EXIT,"wtCDP",0,NULL);
        break;
      } // DEFAULT
    } // SWITCH (iMode)
  } // ELSE

  return 1;

} // cDataPort::writeToCDP()
//==!=====================================================================!==
//==! Convert string to upper or lower case depending on iMode.
//==!
//==! Function Parameters: pStr = pointer to string
//==!                      lMax = maximum string size
//==!                      iMode = 0: convert to lower case
//==!                              1: convert to upper case
//==! Return Value       : 1 if NULL not found, 0 otherwise
//==!=====================================================================!==
int strCaseConvert(char *pStr,long lMax,int iMode)
{ long lX;

  for (lX = 0; pStr[lX] != '\0' && lX < lMax; ++lX)
  { if (isalpha(pStr[lX]))
    { pStr[lX] = (iMode == 0 ? tolower(pStr[lX]) : toupper(pStr[lX]));
    } // IF isalpha()
 } // FOR lX
 if (lX >= lMax)
 { return 1;
 } // IF
 return 0;

} // strCaseConvert()
//==!=====================================================================!==
//==! Message handler.
//==!
//==! Function Parameters: iMsgType = error type
//==!                      sMsg = error message
//==!                      iMode = 0: print message only
//==!                              1: print message and pDP state
//==!                      pDP = data port referred to by sMsg
//==! Return Value       : none
//==!=====================================================================!==
void jMessage(int iMsgType,char *sMsg,int iMode,cDataPort *pDP)
{

  sGlobMessageBuf[0] = '\0';
  if (pDP != NULL)
  { switch(iMsgType)
    { case DISP_SHOW_MSG:
        ++(pDP->iParserWarnings);
        sprintf(sGlobMessageBuf,"Warning: ");
        break;
      case DISP_AND_EXIT:
        sprintf(sGlobMessageBuf,"Error  : ");
        break;
    } // SWITCH (iMsgType)
  } // IF pDP != NULL

  strncat(sGlobMessageBuf,sMsg,MAX_XDATABUF_SIZE-strlen(sGlobMessageBuf)-2);
  dispMsg(sGlobMessageBuf,DISP_POST_LF,1,1);
  if (pDP != NULL && iMode == 1)
  { sprintf(sGlobMessageBuf,"[%s: line %ld]:[%s]",
              pDP->psPortName,pDP->lLineNum,pDP->sOrigData);
    dispMsg(sGlobMessageBuf,DISP_POST_LF,1,1);
  } // IF

  if (iMsgType == DISP_AND_EXIT)
  { t_ProgFinish = time(NULL);
    strcpy(sMsgBuf00,ctime(&t_ProgFinish));
    sprintf(sGlobMessageBuf,"*** Program stopped at %s",sMsgBuf00);
    dispMsg(sGlobMessageBuf,DISP_POST_LF,1,1);
    freeGlobalMem();
    freeApplicationMem();
    exit(EXIT_ERROR);
  } // IF iMsgType == DISP_AND_EXIT

} // jMessage()
//==!=====================================================================!==
//==! Print a file-related error message.
//==!
//==! Function Parameters: iMode = if 0: print message that this file
//==!                                    is missing
//==!                                 1: print message that we cannot open
//==!                                    this file (for read or write)
//==!                                 2: print message as is
//==!                      iFileStatus = type of error for error handler
//==!                      sFileName = filename/filetype being processed
//==! Return Value       : none
//==!=====================================================================!==
void xFileMsg(int iMode,int iFileStatus,char *sFileName)
{
  switch(iMode)
  { case 0:
      sprintf(sMsgOnly,"missing [%s] filename",sFileName);
      break;
    case 1:
      sprintf(sMsgOnly,"(%s) cannot open [%s]",PROGRAM_NAME,sFileName);
      break;
    case 2:
    default:
      sprintf(sMsgOnly,"(%s) %s",PROGRAM_NAME,sFileName);
      break;
  } // SWITCH

  jMessage((iFileStatus == EXIT_ERROR ? DISP_AND_EXIT : DISP_SHOW_MSG),
             sMsgOnly,1,NULL);

} // xFileMsg()
//==!=====================================================================!==
//==! readCombinedOptions() checks each bit position in the long parameter
//==! lOpt to determine option(s) to set.
//==!
//==! Function Parameters: lOpts = logic OR of option settings
//==! Return Value       : none
//==!=====================================================================!==
void readCombinedOptions(unsigned long lOpts)
{
  if (lOpts == 0)   //==! No option bits set
  { return;
  } // IF

  if ((lOpts & OPT_STATUS_OFF) > 0)
  { iOptShowStatus = 0;
  } // IF
  if ((lOpts & OPT_FLOWLOOPDBG) > 0)
  { iOpt_FlowLoopDbg = 1;
  } // IF
  if ((lOpts & OPT_PROGDEBUG) > 0)
  { iOpt_ProgDebug = 1;
  } // IF
  if ((lOpts & OPT_PRT_BSDL) > 0)
  { iOpt_PrintBsdl = 1;
  } // IF
  if ((lOpts & OPT_SKIP_IDCHECK) > 0)
  { iOpt_SkipIdCheck = 1;
  } // IF
  if ((lOpts & OPT_RPTCNTDISPOFF) > 0)
  { iOpt_RptCntDispOff = 1;
  } // IF

} // readCombinedOptions()
//==!=====================================================================!==
//==! Get a number from the cDataPort. Note that what may have started out
//==! as a decimal number (e.g., 3890) may be a variable that starts with a
//==! number (e.g., 411_Call).
//==!
//==! Function Parameters: cChr = first character of potential decimal number
//==!                      pDP = data port pointer
//==! Return Value       : none
//==!=====================================================================!==
void xParseNumData(cDataPort *pDP,char cChr)
{ int iX;

  pDP->sOrigData[0] = cChr;
  pDP->eDataAttr = eATTR_NUM;

  for (iX = 1, getDpChar(pDP,0);
       (iX < MAX_XDATA_SIZE-1) && (pDP->m_iEndOfDataStream == 0);
       ++iX, getDpChar(pDP,0))
  { if (isspace(pDP->iCurChar) ||
        (ispunct(pDP->iCurChar) && pDP->iCurChar != '_'))
    { pDP->iParserGetChar = 0;
      break;
    } // IF
    pDP->sOrigData[iX] = pDP->iCurChar;
    if (isalpha(pDP->iCurChar) || pDP->iCurChar == '_')
    { pDP->eDataAttr = eATTR_VAR;
    } // IF
  } // FOR
  pDP->sOrigData[iX] = '\0';

  if (iX >= MAX_XDATA_SIZE-1)
  { sprintf(sMsgBuf00,"xParseNumData err [%d]",MAX_XDATA_SIZE-1);
    jMessage(DISP_AND_EXIT,sMsgBuf00,1,pDP);
  } // IF

} // xParseNumData()
//==!=====================================================================!==
//==! Check if the character marks the start of a comment in this cDataPort.
//==! The cDataPort member variable iCommentStyle can be set by the
//==! application to control whether it uses the C++ or BSDL/VHDL style
//==! ("--") comments.
//==!
//==! Function Parameters: pDP = data port pointer
//==! Return Value       : 1 if start of a comment, 0 otherwise
//==!=====================================================================!==
int isComment(cDataPort *pDP)
{

  if (pDP->iCommentStyle == pDP->iCurChar)
  { switch(pDP->iCurChar)
    { case XDATA_COMMENT_STYLE_BSDL: // Check for 2 consecutive '-' chars
      { if (pDP->m_iEndOfDataStream == 0)
        { getDpChar(pDP,0);
          if (pDP->iCurChar == XDATA_COMMENT_STYLE_BSDL)
          { pDP->sOrigData[0] = '\0';
            pDP->eDataAttr = eATTR_WSPACE;
            skipDpLine(pDP);
            return 1;
          } // IF
          else
          { pDP->iParserGetChar = 0;
          } // ELSE
        } // IF
        break;
      } // CASE XDATA_COMMENT_STYLE_BSDL

      case XDATA_COMMENT_STYLE_CPP: // Check for 2 consecutive '/' chars
      { if (pDP->m_iEndOfDataStream == 0)
        { getDpChar(pDP,0);
          if (pDP->iCurChar == XDATA_COMMENT_STYLE_CPP)
          { pDP->sOrigData[0] = '\0';
            pDP->eDataAttr = eATTR_WSPACE;
            skipDpLine(pDP);
            return 1;
          } // IF
          else
          { pDP->iParserGetChar = 0;
          } // ELSE
        } // IF
        break;
      } // CASE XDATA_COMMENT_STYLE_CPP

//==! Handle SVF-style comments.
      case XDATA_COMMENT_STYLE_SVF:
      { pDP->sOrigData[0] = '\0';
        pDP->eDataAttr = eATTR_WSPACE;
        skipDpLine(pDP);
        return 1;
      } // CASE XDATA_COMMENT_STYLE_SVF

      default:
      { break;
      } // DEFAULT
    } // SWITCH (pDP->iCurChar)

  } // IF pDP->iCommentStyle == pDP->iCurChar

  return 0;

} // isComment()
//==!=====================================================================!==
//==! Data parser. If iMode:
//==! - XDATA_READ:
//==!   = get non-whitespace data from pDP and store in sOrigData,
//==!     convert data to lowercase and store in sData
//==!   = words/variables are delimited by punctuation marks (except
//==!     underscore) or white-space
//==!   = punctuation marks are delimited by other punctuation marks (except
//==!     underscore), white-space, or the start of a new word/variable
//==!   = ignore white-space
//==! - XDATA_READ_RAW
//==!   = same as XDATA_READ but return white-space
//==!
//==! Function Parameters: pDP = data port pointer
//==!                      iMode = see function description
//==! Return Value       : none
//==!=====================================================================!==
void xParseData(cDataPort *pDP,int iMode)
{ int iX;

  if (pDP->iParserGetData == 1)
  { while (pDP->m_iEndOfDataStream == 0)
    { getDpChar(pDP,0);

//==! Alphanumeric character read
      if (isalnum(pDP->iCurChar) || pDP->iCurChar == '_')
      { if (isdigit(pDP->iCurChar))     //==! data starts with a number
        { xParseNumData(pDP,(char)pDP->iCurChar);
        } // IF
        else
        { for (iX = 0;
               (iX <= MAX_XDATA_SIZE-2) && (pDP->m_iEndOfDataStream == 0);
               ++iX, getDpChar(pDP,0))
          { if (isalnum(pDP->iCurChar) || pDP->iCurChar == '_')
            { pDP->sOrigData[iX] = pDP->iCurChar;
            } // IF
            else
            { break;
            } // ELSE
          } // FOR iX
          pDP->sOrigData[iX] = '\0';
          pDP->eDataAttr = eATTR_VAR;

          if (pDP->m_iEndOfDataStream == 0)
          { pDP->iParserGetChar = 0;
          } // IF

          if (iX >= MAX_XDATA_SIZE-2)
          { sprintf(sMsgOnly,"xParseData bufsize err [%d]",MAX_XDATA_SIZE-2);
            jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
          } // IF
        } // ELSE
      } // IF isalnum()
      else
      { sprintf(pDP->sOrigData,"%c",pDP->iCurChar);
        if (ispunct(pDP->iCurChar))
        { pDP->eDataAttr = eATTR_PUNCT;
          if (isComment(pDP) == 1)
          { continue;
          } // IF
        } // IF ispunct()
        else //==! All other characters treated as white-space.
        { if (isspace(pDP->iCurChar))
          { pDP->eDataAttr = eATTR_WSPACE;
          } // IF
          else
          { pDP->eDataAttr = eATTR_OTHERS;
          } // ELSE
        } // ELSE
      } // ELSE

      strcpy(pDP->sData,pDP->sOrigData);

      if (pDP->eDataAttr != eATTR_WSPACE)
      { if (pDP->eDataAttr == eATTR_VAR)
        { strCaseConvert(pDP->sData,strlen(pDP->sData),0);
        } // IF
        return;
      } // IF
      else
      { if (iMode == XDATA_READ_RAW)
        { return;
        } // IF
      } // ELSE

    } // WHILE pDP->m_iEndOfDataStream == 0
  } // IF iCallParser == 1
  else
  { pDP->iParserGetData = 1;
  } // ELSE

} // xParseData()
//==!=====================================================================!==
//==! Read a long decimal number from the data port.
//==!
//==! Function Parameters: pDP = input data port
//==! Return Value       : long value from data port
//==!=====================================================================!==
long lReadDecimal(cDataPort *pDP)
{ int iSign;
  long lVal;

  xParseData(pDP,XDATA_READ);
  if (xstrcmp(pDP->sData,"-") == 0)
  { iSign = 1;
//==! Number should be right next to '-' sign
    xParseData(pDP,XDATA_READ_RAW);
  } // IF sData == '-'
  else
  { iSign = 0;
  } // ELSE
  if (pDP->eDataAttr != eATTR_NUM)
  { jMessage(DISP_AND_EXIT,"expected a number",1,pDP);
  } // IF
  lVal = atol(pDP->sData);
  if (iSign == 0)
  { return (lVal);
  } // IF
  return (-lVal);

} // lReadDecimal()
//==!=====================================================================!==
//==! Use only in command line mode. Display query/prompt and wait for user's
//==! response. Compare first character of response to array of valid chars
//==! in pValidResp[].
//==!
//==! Function Parameters: pQueryPrompt = prompt
//==!                      pValidResp = char array of valid responses
//==! Return Value       : char from array of valid responses
//==!=====================================================================!==
char xQuery(char *pQueryPrompt,char *pValidResp)
{ char cChr;

  fprintf(stderr,"\n");
  while (1)
  { fprintf(stderr,"%s",pQueryPrompt);
    cChr = toupper(getchar());
    if (pValidResp == NULL)
    { if (cChr == '\n')
      { break;
      } // IF
    } // IF
    while (getchar() != '\n')
      ;
    if (pValidResp == NULL)
    { break;
    } // IF
    if (strchr(pValidResp,cChr))
    { break;
    } // IF
  } // WHILE (1)
  return cChr;

} // xQuery()
//==!=====================================================================!==
//==! Close if pFP != NULL.
//==!
//==! Function Parameters: pFP = pointer to a FILE pointer
//==! Return Value       : none
//==!=====================================================================!==
void xFclose(FILE **pFP)
{

  if (*pFP != NULL)
  { fclose(*pFP);
    *pFP = NULL;
  } // IF

} // xFclose()
//==!=====================================================================!==
//==! Open pFileName using mode "pChkMode":
//==!
//==! - Check mode
//==!   = c  : check if file exists before writing to it
//==!   = cb : same as c but for binary files
//==! - Test mode
//==!   = t  : test if file exists (i.e., return pointer even if NULL)
//==! - Append modes
//==!   = a  : append pFileName
//==!   = ab : same as w but for binary files
//==! - Overwrite modes
//==!   = w  : overwrite pFileName
//==!   = wb : same as w but for binary files
//==! - Read modes
//==!   = r  : open pFileName for read operation
//==!   = rb : same as r but for binary files
//==!
//==! Function Parameters: pFileName = file to open
//==!                      pChkMode = file open mode
//==! Return Value       : FILE pointer to pFileName
//==!=====================================================================!==
FILE *xFileOpen(char *pFileName,char *pChkMode)
{ FILE *pFP;

  switch(pChkMode[0])
  { case 'c':
    case 't':
    { pFP = fopen(pFileName,"r");
      if (pChkMode[0] == 't')
      { return pFP;
      } // IF
      break;
    } // CASE 'c'

    case 'r':
    { pFP = fopen(pFileName,pChkMode);
      if (pFP == NULL)
      { xFileMsg(1,EXIT_ERROR,pFileName);
      } // IF
      return(pFP);
    } // CASE 'r'

    case 'a':
    case 'w':
    { pFP = fopen(pFileName,pChkMode);
      if (pFP == NULL)
      { xFileMsg(1,EXIT_ERROR,pFileName);
      } // IF
      return (pFP);
    } // CASE 'a' or 'w'

    default:
    { sprintf(sMsgBuf00,"Unknown fopen() mode [%s]",pChkMode);
      jMessage(DISP_AND_EXIT,sMsgBuf00,1,NULL);
    } // DEFAULT
  } // SWITCH

  if (pFP != NULL)
  { xFclose(&pFP);
    sprintf(sMsgOnly,"=> Overwrite [%s] or Exit (o/e) ? ",pFileName);
    if (xQuery(sMsgOnly,"OE") == 'E')
    { freeGlobalMem();
      freeApplicationMem();
      exit(EXIT_ERROR);
    } // IF
  } // IF pFP != NULL

//==! Change file open mode from 'c' to 'w'
  pChkMode[0] = 'w';
  pFP = fopen(pFileName,pChkMode);
  if (pFP == NULL)
  { xFileMsg(1,EXIT_ERROR,pFileName);
  } // IF pFP == NULL
  return (pFP);

} // xFileOpen()
//==!====================================================================!==
//==! readCmdLine(): Read/parse the command line. Generic options are:
//==!
//==!   -d = delete output/logfiles without prompting
//==!   -i <input filename> = process an iput file
//==!   -l <log filename> = log operation to specified file
//==!   -o <output filename> = write output to specified file
//==!   -xopt <program option value> = option value is the arithmetic sum of
//==!            different xopt options (search NOTE_0001 for list of valid
//==!            option values)
//==!
//==! Program-specific options are processed by processProgOptions().
//==!
//==! Filenames can have spaces in them when enclosed in double quotes
//==! (e.g., progname -i "strategic account file" -l result.log)
//==!
//==! Function Parameters: argc = argument count
//==!                      argv = pointer to array of char pointers
//==! Return Value       : none
//==!====================================================================!==
void readCmdLine(int argc,char **argv)
{ char *pChr,*pTkn,sFileMode[5];
  FILE *pFP;

  while ((--argc > 0) && ((*++argv)[0] == '-'))
  { pChr = argv[0]+1;
    strCaseConvert(pChr,strlen(pChr),0);

//==! Always overwrite output file(s) =======================================
    if (!strcmp(pChr,"d") || !strcmp(pChr,"w"))
    { iOptOverWritePrompt = 0;
      continue;
    } // IF option = d or w

//==! Check list of input files =============================================
    if (!strcmp(pChr,"i"))
    { freeData(pDP_Input);
      pTkn = readCmdLineData(&argc,&argv,1);
      if (pTkn == NULL)
      { xFileMsg(0,EXIT_ERROR,"InputFile");
      } // IF

      if (pDP_Output != NULL)
      { if (!strcmp(pDP_Output->psPortName,pTkn))
        { xFileMsg(2,EXIT_ERROR,"Input filename same as output");
          break;
        } // IF
      } // IF
      pFP = xFileOpen(pTkn,"r");
      if (pFP == NULL)
      { xFileMsg(0,EXIT_ERROR,"input");
        break;
      } // IF
      xFclose(&pFP);
      pDP_Input = new cDataPort(pTkn);

      continue;
    } // IF option = i

//==! Check log filename ====================================================
    if (!strcmp(pChr,"l"))
    { freeData(pDP_Log);
      pTkn = readCmdLineData(&argc,&argv,1);
      if (pTkn == NULL)
      { xFileMsg(0,EXIT_ERROR,"LogFile");
      } // IF

      sprintf(sFileMode,"%c",'w');
      if (iOptOverWritePrompt == 1)
      { sprintf(sFileMode,"%c",'c');
      } // IF
      pFP = xFileOpen(pTkn,sFileMode);
      if (pFP == NULL)
      { xFileMsg(0,EXIT_ERROR,"logfile");
      } // IF
      pNewDataPort = new cDataPort(pTkn);
      pNewDataPort->pNextDataPort = pDP_List;
      pDP_List = pNewDataPort;
      pDP_Log = pDP_List;
      pDP_Log->pFILE = pFP;
      pFP = NULL;
      continue;
    } // IF option = l

//==! Check output filename =================================================
    if (!strcmp(pChr,"o"))
    { freeData(pDP_Output);
      pTkn = readCmdLineData(&argc,&argv,1);
      if (pTkn == NULL)
      { xFileMsg(0,EXIT_ERROR,"OutFile");
      } // IF

      if (pDP_Input != NULL)
      { if (!strcmp(pDP_Input->psPortName,pTkn))
        { xFileMsg(2,EXIT_ERROR,"Output filename same as input");
          break;
        } // IF
      } // IF
      sprintf(sFileMode,"%c",'w');
      if (iOptOverWritePrompt == 1)
      { sprintf(sFileMode,"%c",'c');
      } // IF
      pFP = xFileOpen(pTkn,sFileMode);
      if (pFP == NULL)
      { xFileMsg(0,EXIT_ERROR,"outfile");
        break;
      } // IF
      xFclose(&pFP);
      pNewDataPort = new cDataPort(pTkn);
      pNewDataPort->pNextDataPort = pDP_List;
      pDP_List = pNewDataPort;
      pDP_Output = pDP_List;
      continue;
    } // IF option = o

//==! Check options and set variables =======================================
    if (!strcmp(pChr,"xopt"))
    { pTkn = readCmdLineData(&argc,&argv,1);
      if (pTkn == NULL)
      { fprintf(stderr,"%s: specify option(s)",PROGRAM_NAME);
        freeGlobalMem();
        freeApplicationMem();
        exit(EXIT_ERROR);
      } // IF
      lSpecOptions = (unsigned long)(atol(pTkn));
      readCombinedOptions(lSpecOptions);
      continue;
    } // IF option = xopt

//==! Check for buffered read operation settings ============================
    if (!strcmp(pChr,"readbufsize"))
    { pTkn = readCmdLineData(&argc,&argv,1);
      if (pTkn == NULL)
      { fprintf(stderr,"%s: specify read buffer size",PROGRAM_NAME);
        freeGlobalMem();
        freeApplicationMem();
        exit(EXIT_ERROR);
      } // IF
      lGlobOptBufferedReadSize = atol(pTkn);
      if (lGlobOptBufferedReadSize < 1024)  // check for minimum buf size
      { lGlobOptBufferedReadSize = dREADBUFMAX;
      } // IF
      continue;
    } // IF option = readbufsize

//==! Check for buffered write operation settings ============================
    if (!strcmp(pChr,"writebufsize"))
    { pTkn = readCmdLineData(&argc,&argv,1);
      if (pTkn == NULL)
      { fprintf(stderr,"%s: specify write buffer size",PROGRAM_NAME);
        freeGlobalMem();
        freeApplicationMem();
        exit(EXIT_ERROR);
      } // IF
      lGlobOptBufferedWriteSize = atol(pTkn);
      if (lGlobOptBufferedWriteSize < 1024)  // check for minimum buf size
      { lGlobOptBufferedWriteSize = dWRITEBUFMAX;
      } // IF
      continue;
    } // IF option = writebufsize

    if (!strcmp(pChr,"nobufread"))
    { iGlobOptBufferedReadOp = 0;
      continue;
    } // IF option = nobufread

    if (!strcmp(pChr,"nobufwrite"))
    { iGlobOptBufferedWriteOp = 0;
      continue;
    } // IF option = nobufwrite

//==! Check other options ===================================================
    if (processProgOptions(&argc,&argv) == 1)
    { continue;
    } // IF

    fprintf(stderr,"%s: unknown option [%s]\n",PROGRAM_NAME,pChr);
    freeGlobalMem();
    freeApplicationMem();
    exit(EXIT_ERROR);
  } // WHILE

//==!========================================================================
//==! Check the following:
//==! - program usually requires at least one input name
//==! - if argc > 0, then command line was incorrectly processed because
//==!   command line arguments still remain.
//==! - if BSDL file checking specified then a command file is not required
//==!   because the BSDL file to check will be specified inthe command line
//==!========================================================================
  if (((pDP_Input == NULL) && (iGlobOptCheckBsdl == 0)) || (argc > 0))
  { iOptShowStatus = 1;
    sprintf(sMsgOnly,"%s %s\n%s",PROG_NAME_TM,VERSION_NUM,PROG_DESC);
    dispMsg(sMsgOnly,DISP_POST_LF,1,1);
    if (argc == 0)
    { dispMsg("==> missing input",DISP_POST_LF,1,1);
    } // IF
    else
    { sprintf(sMsgBuf00,"==> unknown option [%s]",argv[0]);
      dispMsg(sMsgBuf00,DISP_POST_LF,1,1);
    } // ELSE
    sprintf(sMsgBuf00,
       "Usage:\n  %s [options] -i <command file>",
       PROGRAM_NAME);
    dispMsg(sMsgBuf00,DISP_POST_LF,1,1);
    displayProgOptions();
    freeGlobalMem();
    freeApplicationMem();
    exit(EXIT_ERROR);
  } // IF

//==! Check if command line option combinations are valid
  if (checkOptCombo() == 0)  
  { freeGlobalMem();
    freeApplicationMem();
    exit(EXIT_ERROR);
  } // IF

//==! Be sure main() always calls prepareFiles() to open files before calling
//==! the main data/file processing routines.

} // readCmdLine()
//==!=====================================================================!==
//==! Get an extended filename (i.e., can contain spaces or path information)
//==! from pDP and store in pBuf. Filenames containing spaces must be
//==! bracketed by double quotes. If the filename does not contain
//==! whitespace and is not bracketed by double quotes, then the filename
//==! will be delimited by the first whitespace character.
//==!
//==! Function Parameters: pDP = data port pointer
//==!                      pBuf = pointer to char buffer
//==!                      iSize = size of pBuf
//==! Return Value       : none
//==!=====================================================================!==
void getExtFilename(cDataPort *pDP,char *pBuf,int iSize)
{ char *pTmpName;

  pTmpName = NULL;
  pBuf[0] = '\0';
  xParseData(pDP,XDATA_READ);
  if (pDP->sOrigData[0] != '"')
  { do
    { xstrcat(&pTmpName,pDP->sOrigData);
      xParseData(pDP,XDATA_READ_RAW);
    } while (pDP->eDataAttr != eATTR_WSPACE);
  } // IF not '"'
  else
  { while (1)
    { xParseData(pDP,XDATA_READ_RAW);
      if (pDP->sOrigData[0] == '"')
      { break;
      } // IF
      if (pDP->sOrigData[0] == '\n')
      { freeArray(pTmpName);
        jMessage(DISP_AND_EXIT,"Name spans lines",1,pDP);
      } // IF
      xstrcat(&pTmpName,pDP->sOrigData);
    } // WHILE (1)
  } // ELSE

  if ((int)strlen(pTmpName) >= iSize-1)
  { freeArray(pTmpName);
    sprintf(sMsgOnly,"Extfname length [%ld] > [%ld]",strlen(pTmpName),iSize);
    jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
  } // IF
  else
  { strcpy(pBuf,pTmpName);
  } // ELSE
  freeArray(pTmpName);

} // getExtFilename()
//==!=====================================================================!==
//==! Find cChr in the cDataPort.
//==!
//==! Function Parameters: pDP = cDataPort to search
//==!                      cChr = char to find
//==!                      iMode = 0: check next chr only
//==!                              1: continue until cChr found or EOF
//==! Return Value       : 1 if found, 0 otherwise
//==!=====================================================================!==
int findChar(cDataPort *pDP,char cChr,int iMode)
{

  while (pDP->m_iEndOfDataStream == 0)
  { getDpChar(pDP,0);
    if (pDP->iCurChar == cChr)
    { return 1;
    } // IF
    if (iMode == 0)
    { break;
    } // IF iMode == 0
  } // WHILE (pDP->m_iEndOfDataStream == 0)
  sprintf(sMsgOnly,"Char [%c] not found",cChr);
  jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
  return 0;

} // findChar()
//==!=====================================================================!==
//==! Find sStr in the cDataPort.
//==!
//==! Function Parameters: pDP = cDataPort to search
//==!                      sStr = string to find
//==!                      iMode = 0: check next data only
//==!                              1: continue until sStr found or EOF
//==! Return Value       : 1 if found, 0 otherwise
//==!=====================================================================!==
int findStr(cDataPort *pDP,char *sStr,int iMode)
{

  while (pDP->m_iEndOfDataStream == 0)
  { xParseData(pDP,XDATA_READ);
    if (xstrcmp(pDP->sData,sStr) == 0)
    { return 1;
    } // IF
    if (iMode == 0)
    { break;
    } // IF iMode == 0
  } // WHILE (pDP->m_iEndOfDataStream == 0)
  sprintf(sMsgOnly,"[%s] not found",sStr);
  jMessage(DISP_AND_EXIT,sMsgOnly,1,pDP);
  return 0;

} // findStr()
//==!=====================================================================!==
//==! Read a number from cDataPort. It can be a floating number or if
//==! it has no exponent, then it is returned as a long. Could have used
//==! atof() but then will not be able to detect format errors in the
//==! floating point number.
//==!
//==! Function Parameters: pointer to cDataPort
//==!                      pfNum = pointer to floating number
//==!                      plNum = pointer to long number
//==! Return Value       : GF_FLOAT = if floating number was read
//==!                      GF_LONG  = if long integer was read
//==!                      GF_UNKNOWNTYPE = no number was read
//==!=====================================================================!==
int getFloat(cDataPort *pDP,float *fNum,long *lNum)
{ long lMantissa,lExponent;
  int iSign,iSignExponent,iFloat,i,j;

  iSign = 1;  // 1 = positive, 0 = negative
  iSignExponent = 1;
  iFloat = 0;
  lMantissa = 0;
  lExponent = 0;
  *fNum = 0;
  *lNum = 0;
  xParseData(pDP,XDATA_READ);
  switch(pDP->sData[0])
  { case '+':
    case '-':
      if (pDP->sData[0] == '-')
      { iSign = 0;
      } // IF
      xParseData(pDP,XDATA_READ_RAW);
      break;
    case ';':
      return GF_UNKNOWNTYPE;
    default:
      break;
  } // SWITCH

  if (pDP->eDataAttr == eATTR_NUM)
  { lMantissa = atol(pDP->sData);
    xParseData(pDP,XDATA_READ_RAW);
    if (pDP->sData[0] != '.')
    { if (pDP->eDataAttr != eATTR_WSPACE)
      { pDP->iParserGetData = 0;
      } // IF
      *lNum = (iSign == 0 ? -lMantissa : lMantissa);
      if (*lNum < 0)
      { jMessage(DISP_AND_EXIT,GF_POSNUMBER,1,pDP);
      } // IF
      return GF_LONG;
    } // IF
    iFloat = 1;
    xParseData(pDP,XDATA_READ_RAW);
  } // IF token = NUMBER

  if (pDP->eDataAttr != eATTR_NUM && pDP->eDataAttr != eATTR_VAR)
  { jMessage(DISP_AND_EXIT,GF_REALFORMAT,1,pDP);
  } // IF

  for (i = 0; i < (int)strlen(pDP->sData); ++i)
  { if (isalpha(pDP->sData[i]))
    { if (tolower(pDP->sData[i]) != 'e' || i != ((int)strlen(pDP->sData)-1))
      { if (tolower(pDP->sData[i] == 'e'))
        { for (j = i+1; j < (int)strlen(pDP->sData); ++j)
          { if (!isdigit(pDP->sData[j]))
            { break;
            } // IF
          } // FOR j < strlen(sData)
          if (j >= (int)strlen(pDP->sData))
          { strcpy(sMsgBuf00,pDP->sData);
            if (iFloat == 1)
            { sprintf(sMsgBuf00,"%ld.%s",lMantissa,pDP->sData);
            } // IF
            *fNum = (float)atof(sMsgBuf00);
            if (*fNum < 0)
            { jMessage(DISP_AND_EXIT,GF_POSNUMBER,1,pDP);
            } // IF
            return GF_FLOAT;
          } // IF
        } // IF

        if (iFloat == 0)
        { return GF_UNKNOWNTYPE;
        } // IF
        jMessage(DISP_AND_EXIT,GF_REALFORMAT,1,pDP);
      } // IF char != e and index = last position
    } // IF
  } // FOR i < strlen(sData)

  if (iFloat == 1)
  { sprintf(sMsgBuf01,"%ld.%s",lMantissa,pDP->sData);
  } // IF
  else
  { sprintf(sMsgBuf01,"%s",pDP->sData);
  } // ELSE

  xParseData(pDP,XDATA_READ_RAW);
  if (pDP->eDataAttr != eATTR_VAR)
  { switch(pDP->sData[0])
    { case '+':
        break;
      case '-':
        iSignExponent = 0;
        break;
      default:
        jMessage(DISP_AND_EXIT,"Exponent sign ?",1,pDP);
        break;
    } // SWITCH
    lExponent = lReadDecimal(pDP);
  } // IF
  else
  { pDP->iParserGetData = 0;
  } // ELSE

  sprintf(sMsgBuf00,"%c%s%c%ld",(iSign == 1 ? ' ' : '-'),
          sMsgBuf01,(iSignExponent == 1 ? '+' : '-'),
          lExponent);
  *fNum = (float)atof(sMsgBuf00);
  if (*fNum < 0)
  { jMessage(DISP_AND_EXIT,GF_POSNUMBER,1,pDP);
  } // IF
  return GF_FLOAT;

} // getFloat()
//==!=====================================================================!==
//==! Case-insensitive string comparison routine. Use this instead of using
//==! stricmp() which works only on Windows.
//==!
//==! Function Parameters: sA = pointer str A
//==!                      sB = pointer str B
//==! Return Value       : 0 if equal, difference in chars otherwise
//==!=====================================================================!==
int xstrcmp(char *sA,char *sB)
{ int iVar,iA,iB;

  iA = strlen(sA);
  iB = strlen(sB);
  iVar = iA - iB;
  if (iVar != 0)
  { return iVar;
  } // IF
  for (iVar = 0; iVar < iA; ++iVar)
  { if (tolower(sA[iVar]) - tolower(sB[iVar]) != 0)
    { return (tolower(sA[iVar]) - tolower(sB[iVar]));
    } // IF
  } // FOR iVar < iA
  return 0;

} // xstrcmp()
//==!=====================================================================!==
//==! Concatenation routine: concatenates sSrc to sDest by allocating the
//==! appropriate amount of memory. Developer could have used MFC class
//==! CString but this may not be portable to Unix.
//==!
//==! sDest must be a pointer to a (char *) because its size will be modified.
//==!
//==! Function Parameters: sDest = pointer to destination string pointer
//==!                      sSrc = source string pointer
//==! Return Value       : pointer to CTapField object
//==!=====================================================================!==
void xstrcat(char **sDest,char *sSrc)
{ long lCnt;
  char *sTmp;

  lCnt = ((*sDest) == NULL ? 0 : (long)strlen(*sDest));
  lCnt += (long)strlen(sSrc) + 1;
  sTmp = new char [lCnt];
  if (*sDest == NULL)
  { sprintf(sTmp,"%s",sSrc);
  } // IF
  else
  { sprintf(sTmp,"%s%s",*sDest,sSrc);
    delete [] (*sDest);
  } // ELSE
  *sDest = sTmp;

} // xstrcat()
//==!=====================================================================!==
//==! Initialize array cBuf[0..iCnt-1] with cChar.
//==!
//==! Function Parameters: cBuf = pointer to char array
//==!                      iCnt = number of array locations to fill
//==!                      cFill = fill character
//==!                      iMODE = 0: do not put a NULL at end of iCnt
//==!                              1: terminate array with a NULL
//==! Return Value       : none
//==!=====================================================================!==
void xInitCharArray(char *cBuf,int iCnt,char cFill,int iMode)
{ int x;

  for (x = 0; x < iCnt; ++x)
  { cBuf[x] = cFill;
  } // FOR x
  if (iMode == 1)
  { cBuf[x] = '\0';
  } // IF

} // xInitCharArray()
//==!=====================================================================!==
//==! end of module XPARSER.CPP
//==!=====================================================================!==

